/**
 * EasyBeans
 * Copyright (C) 2006-2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MetadataMerge.java 4121 2008-10-06 08:21:14Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.common.impl.xml.merge;

import org.ow2.util.ee.metadata.common.api.ICommonClassMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonFieldMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonMethodMetadata;
import org.ow2.util.ee.metadata.common.api.interfaces.IEnvironmentRefAccessor;
import org.ow2.util.ee.metadata.common.api.interfaces.ISharedMetadata;
import org.ow2.util.ee.metadata.common.api.struct.IJAnnotationResource;
import org.ow2.util.ee.metadata.common.api.struct.IJEjbEJB;
import org.ow2.util.ee.metadata.common.api.struct.IJaxwsWebServiceRef;
import org.ow2.util.ee.metadata.common.api.struct.IEnvEntry;
import org.ow2.util.ee.metadata.common.impl.struct.JAnnotationResource;
import org.ow2.util.ee.metadata.common.impl.struct.JEjbEJB;
import org.ow2.util.ee.metadata.common.impl.struct.JaxwsWebServiceRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.AbsEJBRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.AbsEnvironment;
import org.ow2.util.ee.metadata.common.impl.xml.struct.AroundInvoke;
import org.ow2.util.ee.metadata.common.impl.xml.struct.EJBLocalRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.EJBRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.InjectionTarget;
import org.ow2.util.ee.metadata.common.impl.xml.struct.LifeCycleCallback;
import static org.ow2.util.ee.metadata.common.impl.xml.struct.LifeCycleCallback.POST_CONSTRUCT;
import static org.ow2.util.ee.metadata.common.impl.xml.struct.LifeCycleCallback.PRE_DESTROY;
import org.ow2.util.ee.metadata.common.impl.xml.struct.MessageDestinationRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.ResourceEnvRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.ResourceRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.ServiceRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.EnvEntry;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * Perform a common merge from XML files to metadata.
 * @author Gael Lalire
 */
public abstract class MetadataMerge {

    /**
     * Defines java.lang.Object class.
     */
    public static final String JAVA_LANG_OBJECT = "java/lang/Object";

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(MetadataMerge.class);

    /**
     * Apply rules for a an environment.
     * @param envGroup the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    protected void applyJndiEnvironmentRefsGroup(final AbsEnvironment envGroup,
            final IEnvironmentRefAccessor classAnnotationMetadata) {

        // post-construct / pre-destroy
        applyLifeCycle(envGroup.getPostConstructCallbackList(), classAnnotationMetadata, POST_CONSTRUCT);
        applyLifeCycle(envGroup.getPreDestroyCallbackList(), classAnnotationMetadata, PRE_DESTROY);

        // ejb(local) ref
        applyEjbRef(envGroup.getEJBRefList(), classAnnotationMetadata);
        applyEjbLocalRef(envGroup.getEJBLocalRefList(), classAnnotationMetadata);

        // resource(env) ref
        applyResourceRef(envGroup.getResourceRefList(), classAnnotationMetadata);
        applyResourceEnvRef(envGroup.getResourceEnvRefList(), classAnnotationMetadata);

        // message-destination-ref
        applyMessageDestinationRef(envGroup.getMessageDestinationRefList(), classAnnotationMetadata);

        // env entry
       applyEnvEntry(envGroup.getEnvEntryList(), classAnnotationMetadata);


        /* in ejb jar
         *         applyLifeCycle(envGroup.getPrePassivateCallbackList(), classAnnotationMetadata, PRE_PASSIVATE);
        applyLifeCycle(envGroup.getPostActivateCallbackList(), classAnnotationMetadata, POST_ACTIVATE);
        applyAroundInvoke(envGroup.getAroundInvokeList(), classAnnotationMetadata, AroundInvoke.NAME);

         */

        // service-ref
        applyServiceRef(envGroup.getServiceRefList(), classAnnotationMetadata);

    }

    /**
     * Apply rules for a common bean (service-ref).
     * @param serviceRefList the list of service-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    protected void applyServiceRef(final List<ServiceRef> serviceRefList,
                                   final IEnvironmentRefAccessor classAnnotationMetadata) {
        if (serviceRefList == null) {
            return;
        }

        // For each service-ref, build a &#64;{@link
        // javax.xml.ws.WebServiceRef}
        for (ServiceRef serviceRef : serviceRefList) {

            List<InjectionTarget> injectionTargetList = serviceRef.getInjectionTargetList();

            // Honor the injection-target if specified
            if (injectionTargetList != null) {
                for (InjectionTarget target : injectionTargetList) {
                    // Find the attribute / target
                    ISharedMetadata sharedMetadata = getInjectionTarget(classAnnotationMetadata, target);

                    // If there is already a WebServiceRef annotation, overload it
                    IJaxwsWebServiceRef wsr = sharedMetadata.getJaxwsWebServiceRef();

                    if (wsr != null) {
                        // merge
                        mergeWebServiceRef(serviceRef, wsr);
                    } else {
                        // place a new annotation on the target
                        sharedMetadata.setJaxwsWebServiceRef(createJaxwsWebServiceRefFromXML(serviceRef));
                    }
                }
            } else {
                // We need to lookup the target annotation in the class
                // and overload it.
                String name = serviceRef.getName();
                List<IJaxwsWebServiceRef> metadatas = findJaxwsWebServiceRefMetadata(classAnnotationMetadata, name);

                if (metadatas.isEmpty()) {
                    // no annotation to override:
                    // -> equivalent to class-level annotation

                    // If there is an existing single class-leve annotation, place it in the list
                    IJaxwsWebServiceRef existing = classAnnotationMetadata.getJaxwsWebServiceRef();
                    List<IJaxwsWebServiceRef> webServiceRefs = classAnnotationMetadata.getJaxwsWebServiceRefs();

                    // init the list if needed
                    if (webServiceRefs == null) {
                        webServiceRefs = new ArrayList<IJaxwsWebServiceRef>();
                        classAnnotationMetadata.setJaxwsWebServiceRefs(webServiceRefs);
                    }

                    if (existing != null) {
                        // place the annotation in the @WebServiceRefs list
                        webServiceRefs.add(existing);
                        // remove the single @WebServiceRef
                        classAnnotationMetadata.setJaxwsWebServiceRef(null);
                    }

                    // add our new annotation in the list
                    webServiceRefs.add(createJaxwsWebServiceRefFromXML(serviceRef));


                } else if (metadatas.size() > 1) {
                    // Not good, too many choices
                    // a smart tool may try to find a compatible existing metadata and perform the merge
                    // and a dumb tool will produce an Exception :)
                    throw new IllegalStateException("Too many matching WebServiceRef metadata for '" + name
                                                    + "' service-ref of class '" + classAnnotationMetadata.getEnvironmentName() + "'");
                } else {
                    // only 1 matching candidate
                    // merge
                    mergeWebServiceRef(serviceRef, metadatas.get(0));
                }
            }
        }
    }

    private List<IJaxwsWebServiceRef> findJaxwsWebServiceRefMetadata(final IEnvironmentRefAccessor classAnnotationMetadata,
                                                                     final String name) {

        List<IJaxwsWebServiceRef> elements = new ArrayList<IJaxwsWebServiceRef>();

        // Find annotation on Class
        // Try @WebServiceRef
        IJaxwsWebServiceRef classAnnotation = classAnnotationMetadata.getJaxwsWebServiceRef();
        if ((classAnnotation != null) && (name.equals(classAnnotation.getName()))) {
            elements.add(classAnnotation);
        }
        // Try @WebServiceRefs
        List<IJaxwsWebServiceRef> webServiceRefs = classAnnotationMetadata.getJaxwsWebServiceRefs();
        if (webServiceRefs != null) {
            for (IJaxwsWebServiceRef ref : webServiceRefs) {
                if (name.equals(ref.getName())) {
                    elements.add(classAnnotation);
                }
            }
        }


        /*
        // Find annotation on methods
        for (ICommonMethodMetadata<?, ?, ?> methodMetadata : classAnnotationMetadata.getStandardMethodMetadataCollection()) {
            IJaxwsWebServiceRef methodAnnotation = methodMetadata.getJaxwsWebServiceRef();
            if ((methodAnnotation != null) && (name.equals(methodAnnotation.getName()))) {
                elements.add(methodAnnotation);
            }
        }

        // Find annotation on fields
        for (ICommonFieldMetadata<?, ?, ?> fieldMetadata : classAnnotationMetadata.getStandardFieldMetadataCollection()) {
            IJaxwsWebServiceRef methodAnnotation = fieldMetadata.getJaxwsWebServiceRef();
            if ((methodAnnotation != null) && (name.equals(methodAnnotation.getName()))) {
                elements.add(methodAnnotation);
            }
        }*/

        return elements;
    }

    private IJaxwsWebServiceRef createJaxwsWebServiceRefFromXML(final ServiceRef ref) {
        // Creates an empty metadata element
        IJaxwsWebServiceRef metadata = new JaxwsWebServiceRef();

        // Fill it from XML values
        metadata.setName(ref.getName());
        metadata.setWsdlLocation(ref.getWsdlFile().toString());

        // service-ref-type
        metadata.setType(ref.getServiceRefType());

        // service-interface
        metadata.setValue(ref.getServiceInterface());

        // TODO to be completed: service-qname + ...

        // Return
        return metadata;
    }

    /**
     * Merge the values from the XML inside the metadata struct.
     * @param serviceRef XML
     * @param annotation the metadata structure to be updated
     */
    private void mergeWebServiceRef(final ServiceRef serviceRef,
                                    final IJaxwsWebServiceRef annotation) {

        if (serviceRef.getWsdlFile() != null) {
            annotation.setWsdlLocation(serviceRef.getWsdlFile().toString());
        }

        if (serviceRef.getServiceInterface() != null) {
            if (annotation.getValue() == null) {
                // the annotation is used to declare a service
                annotation.setType(serviceRef.getServiceInterface());
            } else if (annotation.getType() == null) {
                // TODO check if the type is a Service subclass or not
                // the annotation is used to declare a service-endpoint
                annotation.setValue(serviceRef.getServiceInterface());
            }
        }

        if (serviceRef.getServiceRefType() != null) {
            annotation.setType(serviceRef.getServiceRefType());
        }

        // Override name
        annotation.setName(serviceRef.getName());

        // TODO complete merging (port-component-ref + serice-qname)
    }

    /**
     * Apply rules for a common bean (message-destination-ref).
     * @param messageDestinationRefList the list of message-destination-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    protected void applyMessageDestinationRef(final List<MessageDestinationRef> messageDestinationRefList,
            final IEnvironmentRefAccessor classAnnotationMetadata) {
        if (messageDestinationRefList == null) {
            return;
        }

        // For each message-destination-ref, build a &#64;{@link
        // javax.annotation.Resource}
        for (MessageDestinationRef messageDestinationRef : messageDestinationRefList) {
            JAnnotationResource jResource = new JAnnotationResource();
            // set name
            jResource.setName(messageDestinationRef.getName());

            // set type
            jResource.setType(messageDestinationRef.getType());

            // Mapped Name
            jResource.setMappedName(messageDestinationRef.getMappedName());

            // Apply the resource on the class
            applyJResource(jResource, classAnnotationMetadata, messageDestinationRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (resource-ref).
     * @param resourceRefList the list of resource-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    protected void applyResourceRef(final List<ResourceRef> resourceRefList,
            final IEnvironmentRefAccessor classAnnotationMetadata) {
        if (resourceRefList == null) {
            return;
        }

        // For each resource-ref, build a &#64;{@link javax.annotation.Resource}
        for (ResourceRef resourceRef : resourceRefList) {
            JAnnotationResource jResource = new JAnnotationResource();
            // set name
            jResource.setName(resourceRef.getResRefName());

            // set type
            jResource.setType(resourceRef.getResRefType());

            // Mapped Name
            jResource.setMappedName(resourceRef.getMappedName());

            // Apply the resource on the class
            applyJResource(jResource, classAnnotationMetadata, resourceRef.getInjectionTargetList());
        }
    }


    /**
     * Apply rules for a common bean (resource-env-ref).
     * @param resourceEnvRefList the list of resource-env-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    protected void applyResourceEnvRef(final List<ResourceEnvRef> resourceEnvRefList,
            final IEnvironmentRefAccessor classAnnotationMetadata) {
        if (resourceEnvRefList == null) {
            return;
        }

        // For each resource-ref, build a &#64;{@link javax.annotation.Resource}
        for (ResourceEnvRef resourceEnvRef : resourceEnvRefList) {
            JAnnotationResource jResource = new JAnnotationResource();
            // set name
            jResource.setName(resourceEnvRef.getResourceEnvRefName());

            // set type
            jResource.setType(resourceEnvRef.getResourceEnvRefType());

            // Mapped Name
            jResource.setMappedName(resourceEnvRef.getMappedName());

            // Apply the resource on the class
            applyJResource(jResource, classAnnotationMetadata, resourceEnvRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (env-entry).
     * @param envEntryList the list of env-entry.
     * @param classAnnotationMetadata (Annotation metadata) of the bean
     */
    protected void applyEnvEntry(final List<EnvEntry> envEntryList,
            final IEnvironmentRefAccessor classAnnotationMetadata) {
        if (envEntryList == null) {
            return;
        }

        List<IEnvEntry> structEnvEntryList = new ArrayList<IEnvEntry>();

        // Store the env-entry list in the metadata class
        for (EnvEntry envEntry : envEntryList) {
            IEnvEntry builtEntry = new  org.ow2.util.ee.metadata.common.impl.struct.EnvEntry(envEntry.getEnvEntryName(), envEntry
                    .getEnvEntryType(), envEntry.getEnvEntryValue());
            structEnvEntryList.add(builtEntry);

            // for each injection-target field, add a resource annotation
            List<InjectionTarget> injectionTargetList = envEntry.getInjectionTargetList();
            if (injectionTargetList != null) {
                for (InjectionTarget injectionTarget : injectionTargetList) {

                    // build a new annotation
                    JAnnotationResource jResource = new JAnnotationResource();

                    // set name
                    jResource.setName(envEntry.getEnvEntryName());

                    // set type
                    jResource.setType(envEntry.getEnvEntryType());

                    // Add the JAnnotationResource on it
                    getInjectionTarget(classAnnotationMetadata, injectionTarget).setJAnnotationResource(jResource);
                }
            }



            //applyJResource(jResource, classAnnotationMetadata, resourceEnvRef.getInjectionTargetList());
        }

        // Set the env entry collection
        classAnnotationMetadata.setEnvEntryCollection(structEnvEntryList);

    }

    /**
     * Apply rules for a common bean (ejb-ref).
     * @param ejbRefList the list of ejb-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    protected void applyEjbRef(final List<EJBRef> ejbRefList, final IEnvironmentRefAccessor classAnnotationMetadata) {
        if (ejbRefList == null) {
            return;
        }

        // For each ejb-ref, build a &#64;{@link javax.ejb.EJB}
        for (EJBRef ejbRef : ejbRefList) {
            JEjbEJB jEJB = buildEJB(ejbRef);
            // set interface
            if (ejbRef.getHome() != null) {
                jEJB.setBeanInterface(ejbRef.getHome());
            } else if (ejbRef.getRemote() != null) {
                jEJB.setBeanInterface(ejbRef.getRemote());
            }
            applyJEJB(jEJB, classAnnotationMetadata, ejbRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (ejb-local-ref).
     * @param ejbLocalRefList the list of ejb-local-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    protected void applyEjbLocalRef(final List<EJBLocalRef> ejbLocalRefList,
            final IEnvironmentRefAccessor classAnnotationMetadata) {
        if (ejbLocalRefList == null) {
            return;
        }

        // For each ejb-local-ref, build a &#64;{@link javax.ejb.EJB}
        for (EJBLocalRef ejbLocalRef : ejbLocalRefList) {
            JEjbEJB jEJB = buildEJB(ejbLocalRef);
            // set interface
            if (ejbLocalRef.getLocalHome() != null) {
                jEJB.setBeanInterface(ejbLocalRef.getLocalHome());
            } else if (ejbLocalRef.getLocal() != null) {
                jEJB.setBeanInterface(ejbLocalRef.getLocal());
            }
            applyJEJB(jEJB, classAnnotationMetadata, ejbLocalRef.getInjectionTargetList());
        }
    }

    /**
     * Adds the given &#64;{@link javax.annotation.Resource} object to the
     * class and apply it also on injection target elements.
     * @param jResource an &#64;{@link javax.annotation.Resource} impl
     * @param classAnnotationMetadata the bean's class
     * @param injectionTargetList the list of injection fields to set
     */
    private void applyJResource(final JAnnotationResource jResource,
            final IEnvironmentRefAccessor classAnnotationMetadata, final List<InjectionTarget> injectionTargetList) {

        // Add field on the class
        IJAnnotationResource existingJResource = classAnnotationMetadata.getJAnnotationResource();

        // Needs to use a list instead.
        List<IJAnnotationResource> jResources = classAnnotationMetadata.getJAnnotationResources();
        if (jResources == null) {
            jResources = new ArrayList<IJAnnotationResource>();
            classAnnotationMetadata.setJAnnotationResources(jResources);
        }

        // put existing element in the list
        if (existingJResource != null) {
            jResources.add(existingJResource);
            // remove single element
            classAnnotationMetadata.setJAnnotationResource(null);
        }

        // Add our new element
        jResources.add(jResource);

        // Now, add elements on the injection fields
        if (injectionTargetList != null) {
            for (InjectionTarget injectionTarget : injectionTargetList) {

                // Add the JAnnotationResource on it
                getInjectionTarget(classAnnotationMetadata, injectionTarget).setJAnnotationResource(jResource);
            }
        }

    }

    /**
     * Gets the metadata for a given injection target.
     * @param classAnnotationMetadata the class used to search the target
     *        element
     * @param injectionTarget the given target info element
     * @return a common annotation metadata if element is found
     */
    private ISharedMetadata getInjectionTarget(final IEnvironmentRefAccessor classAnnotationMetadata,
            final InjectionTarget injectionTarget) {
        String injectionTargetName = injectionTarget.getTargetName();

        // First, try with the method
        ICommonMethodMetadata<?, ?, ?> method = getMethod(classAnnotationMetadata, injectionTarget);

        // found ? return it
        if (method != null) {
            return method;
        }

        // else, try with the next item which should be a field
        // Try with the field
        ICommonFieldMetadata<?, ?, ?> field = getField(classAnnotationMetadata, injectionTarget);
        if (field == null) {
            throw new IllegalArgumentException("No method or field for injection target name '" + injectionTargetName
                    + "' found in the class '" + classAnnotationMetadata.getEnvironmentName() + "'.");
        }

        return field;

    }

    /**
     * Gets the method metadata for a given injection target.
     * @param classAnnotationMetadata the class used to search the target
     *        element
     * @param injectionTarget the given target info element
     * @return a method metadata if element is found
     */
    private ICommonMethodMetadata<?, ?, ?> getMethod(final IEnvironmentRefAccessor classAnnotationMetadata,
            final InjectionTarget injectionTarget) {

        // get data from the injection target element
        String clName = injectionTarget.getClassname();

        if (clName == null) {
            clName = findDefaultClassForEnvironment(classAnnotationMetadata);
        } else {
            clName = clName.replace(".", "/");
        }

        // Get method name which is the injection target name
        String methodName = injectionTarget.getTargetName();

        // Now, search the element in the given class
        ICommonClassMetadata<?, ?, ?> foundClassMetadata = getClassMetadata(clName);
        if (foundClassMetadata == null) {
            throw new IllegalArgumentException("Cannot find the class metadata named '" + clName + "'.");
        }

        // Got the class metadata, now search the field
        List<? extends ICommonMethodMetadata<?, ?, ?>> methods = foundClassMetadata.searchStandardMethodMetadata(methodName);

        if (methods.size() == 0) {
            // no method
            return null;
        } else if (methods.size() > 1) {
            // Should be only one
            throw new IllegalArgumentException("Too many methods with name '" + methodName + "' found in the class '" + clName
                    + "'.");
        }

        return methods.get(0);
    }

    /**
     * Gets the field metadata for a given injection target.
     * @param classAnnotationMetadata the class used to search the target
     *        element
     * @param injectionTarget the given target info element
     * @return a field if element is found
     */
    private ICommonFieldMetadata<?, ?, ?> getField(final IEnvironmentRefAccessor classAnnotationMetadata,
            final InjectionTarget injectionTarget) {
        // get data from the injection target element
        String clName = injectionTarget.getClassname();
        if (clName == null) {
            clName = findDefaultClassForEnvironment(classAnnotationMetadata);
        } else {
            clName = clName.replace(".", "/");
        }

        String fieldName = injectionTarget.getTargetName();

        // Now, search the element in the given class
        ICommonClassMetadata<?, ?, ?> foundClassMetadata = getClassMetadata(clName);
        if (foundClassMetadata == null) {
            throw new IllegalArgumentException("Cannot find the class metadata named '" + clName + "'.");
        }

        // Got the class metadata, now search the field
        List<? extends ICommonFieldMetadata<?, ?, ?>> fields = foundClassMetadata.searchStandardFieldMetadata(fieldName);

        if (fields.size() == 0) {
            // no field
            return null;
        } else if (fields.size() > 1) {
            // Should be only one
            throw new IllegalArgumentException("Too many fields with name '" + fieldName + "' found in the class '" + clName
                    + "'.");
        }

        return fields.get(0);
    }

    /**
     * Adds the given &#64;{@link javax.ejb.EJB} object to the class and apply
     * it also on injection target elements.
     * @param jEJB an &#64;{@link javax.ejb.EJB} impl
     * @param classAnnotationMetadata the bean's class
     * @param injectionTargetList the list of injection fields to set
     */
    private void applyJEJB(final JEjbEJB jEJB, final IEnvironmentRefAccessor classAnnotationMetadata,
            final List<InjectionTarget> injectionTargetList) {

        // Add field on the class
        IJEjbEJB existingJEjbEJB = classAnnotationMetadata.getJEjbEJB();

        // Needs to use a list instead.
        List<IJEjbEJB> jEjbEJBs = classAnnotationMetadata.getJEjbEJBs();
        if (jEjbEJBs == null) {
            jEjbEJBs = new ArrayList<IJEjbEJB>();
            classAnnotationMetadata.setJEjbEJBs(jEjbEJBs);
        }

        // put existing element in the list
        if (existingJEjbEJB != null) {
            jEjbEJBs.add(existingJEjbEJB);
            // remove single element
            classAnnotationMetadata.setJEjbEJB(null);
        }

        // Add our new element
        jEjbEJBs.add(jEJB);

        // Now, add elements on the injection fields
        if (injectionTargetList != null) {
            for (InjectionTarget injectionTarget : injectionTargetList) {

                // Add the JEjbEJB on it
                getInjectionTarget(classAnnotationMetadata, injectionTarget).setJEjbEJB(jEJB);
            }
        }

    }

    /**
     * Build an &#64;{@link javax.ejb.EJB} impl for the given
     * ejb-ref/ejb-local-ref.
     * @param ref an instance of a ejb-ref/ejb-local-ref.
     * @return a implementation of &#64;{@link javax.ejb.EJB} interface.
     */
    private JEjbEJB buildEJB(final AbsEJBRef ref) {
        JEjbEJB jEJB = new JEjbEJB();
        // Sets the name
        if (ref.getEjbRefName() != null) {
            jEJB.setName(ref.getEjbRefName());
        }

        // sets the mapped-name
        if (ref.getMappedName() != null) {
            jEJB.setMappedName(ref.getMappedName());
        }

        // sets the bean-name
        if (ref.getEjbLink() != null) {
            jEJB.setBeanName(ref.getEjbLink());
        }

        return jEJB;
    }

    /**
     * Apply rules for a common bean (method callbacks).
     * @param aroundInvokeList the list of method/aroundInvoke callbacks.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     * @param type the type of lifecycle.
     */
    protected void applyAroundInvoke(final List<AroundInvoke> aroundInvokeList,
            final IEnvironmentRefAccessor classAnnotationMetadata, final String type) {
        for (AroundInvoke aroundInvoke : aroundInvokeList) {
            String interceptorClassName = null;

            // if no class, takes the bean's class
            if (aroundInvoke.getClassName() == null) {
                interceptorClassName = findDefaultClassForEnvironment(classAnnotationMetadata);
            } else {
                interceptorClassName = encode(aroundInvoke.getClassName());
            }

            // Now, search the Method on the class or its super-class if not
            // found
            String searchMethodName = aroundInvoke.getMethodName();
            if (searchMethodName == null) {
                throw new IllegalStateException("Invalid null method name for the callback '" + interceptorClassName + "'.");
            }

            // Apply interceptor
            applyInterceptor(type, interceptorClassName, searchMethodName, classAnnotationMetadata);

        }
    }

    /**
     * Apply rules for a common bean (lifecycle callbacks).
     * @param lifecycleList the list of lifecycle callbacks.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     * @param type the type of lifecycle.
     */
    protected void applyLifeCycle(final List<LifeCycleCallback> lifecycleList,
            final IEnvironmentRefAccessor classAnnotationMetadata, final String type) {

        if (lifecycleList == null) {
            return;
        }

        for (LifeCycleCallback lifecycleCallback : lifecycleList) {

            String interceptorClassName = null;

            // if no class, takes the bean's class
            if (lifecycleCallback.getLifecycleCallbackClass() == null) {
                interceptorClassName = findDefaultClassForEnvironment(classAnnotationMetadata);
            } else {
                interceptorClassName = encode(lifecycleCallback.getLifecycleCallbackClass());
            }

            // Now, search the Method on the class or its super-class if not
            // found
            String searchMethodName = lifecycleCallback.getLifecycleCallbackMethod();
            if (searchMethodName == null) {
                throw new IllegalStateException("Invalid null method name for the callback '" + interceptorClassName + "'.");
            }

            // Apply interceptor
            applyInterceptor(type, interceptorClassName, searchMethodName, classAnnotationMetadata);

        }
    }


    /**
     * Finds a method in the interceptor class and set the correct type value.
     * @param type type of interceptor
     * @param interceptorClassName the name of the interceptor class
     * @param methodName the name of the method in the interceptor class
     * @param classAnnotationMetadata the class that is using the interceptor
     *        (may be null)
     */
    private void applyInterceptor(final String type, final String interceptorClassName, final String methodName,
            final IEnvironmentRefAccessor classAnnotationMetadata) {

        // if the bean is using this interceptor and that this is not the bean
        // class, add the link.
        if (classAnnotationMetadata != null && !interceptorClassName.equals(findDefaultClassForEnvironment(classAnnotationMetadata))) {
          // 111  addInterceptorForClass(classAnnotationMetadata, interceptorClassName);
        }

        ICommonMethodMetadata<?, ?, ?> foundMethod = null;
        ICommonClassMetadata<?, ?, ?> interceptorMetadata = getClassMetadata(interceptorClassName);

        boolean loop = true;
        // While method is not found and that class has super class, try to
        // search
        while (foundMethod == null && loop) {
            List<? extends ICommonMethodMetadata<?, ?, ?>> listMethods = interceptorMetadata
                    .searchStandardMethodMetadata(methodName);

            // Method present ?
            if (listMethods.size() == 1) {
                // first item
                foundMethod = listMethods.get(0);
                continue;
            } else if (listMethods.size() > 1) {
                throw new IllegalStateException("Method with name '" + methodName + "' was found more than once on the class '"
                        + interceptorMetadata.getJClass().getName() + "'.");
            }

            // else continue
            String superClassName = interceptorMetadata.getJClass().getSuperName();
            if (!superClassName.equals(JAVA_LANG_OBJECT)) {
                interceptorMetadata = getClassMetadata(superClassName);
            } else {
                // break loop
                loop = false;
            }
        }

        if (foundMethod == null) {
            throw new IllegalStateException("The method named '" + methodName + "' was not found in the class '"
                    + interceptorMetadata.getJClass().getName() + "' and all its super-classes.");
        }

        // call setter to true and only if it is not already present.
        if (POST_CONSTRUCT.equals(type)) {
            if (!foundMethod.isPostConstruct()) {
                foundMethod.setPostConstruct(true);
             // TODO gaellalire : move to ezb -> interceptorMetadata.addPostConstructMethodMetadata(foundMethod);
            }
        } else if (PRE_DESTROY.equals(type)) {
            if (!foundMethod.isPreDestroy()) {
                foundMethod.setPreDestroy(true);
             // TODO gaellalire : move to ezb -> interceptorMetadata.addPreDestroyMethodMetadata(foundMethod);
            }
        }
        /* 111
        else if (PRE_PASSIVATE.equals(type)) {
            if (!foundMethod.isPrePassivate()) {
                foundMethod.setPrePassivate(true);
             // TODO gaellalire : move to ezb -> interceptorMetadata.addPrePassivateMethodMetadata(foundMethod);
            }
        } else if (POST_ACTIVATE.equals(type)) {
            if (!foundMethod.isPostActivate()) {
                foundMethod.setPostActivate(true);
             // TODO gaellalire : move to ezb -> interceptorMetadata.addPostActivateMethodMetadata(foundMethod);
            }
        } else if (AroundInvoke.NAME.equals(type)) {
            if (!foundMethod.isAroundInvoke()) {
                foundMethod.setAroundInvoke(true);
             // TODO gaellalire : move to ezb -> interceptorMetadata.addAroundInvokeMethodMetadata(foundMethod);
            }
        } else {
            throw new IllegalStateException("No case for type '" + type + "'..");
        }
        */
    }

    /**
     * Encode a classname (replace . by /) as it's the internal form of the
     * metadata
     * @param className the name of the class to encode
     * @return encoded name
     */
    protected static String encode(final String className) {
        return className.replace(".", "/");
    }

    /**
     * @param className the class name
     * @return the common class metadata which name is classname
     */
    public abstract ICommonClassMetadata<?, ?, ?> getClassMetadata(String className);

    public abstract String findDefaultClassForEnvironment(IEnvironmentRefAccessor env);

}
