/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: CommonDeploymentDescLoader.java 4121 2008-10-06 08:21:14Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.common.impl.xml.parsing;

import org.ow2.util.ee.metadata.common.impl.xml.struct.AbsEJBRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.AbsEnvironment;
import org.ow2.util.ee.metadata.common.impl.xml.struct.AbsResourceGroup;
import org.ow2.util.ee.metadata.common.impl.xml.struct.EJBLocalRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.EJBRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.EnvEntry;
import org.ow2.util.ee.metadata.common.impl.xml.struct.InjectionTarget;
import org.ow2.util.ee.metadata.common.impl.xml.struct.LifeCycleCallback;
import static org.ow2.util.ee.metadata.common.impl.xml.struct.LifeCycleCallback.POST_ACTIVATE;
import static org.ow2.util.ee.metadata.common.impl.xml.struct.LifeCycleCallback.POST_CONSTRUCT;
import static org.ow2.util.ee.metadata.common.impl.xml.struct.LifeCycleCallback.PRE_DESTROY;
import static org.ow2.util.ee.metadata.common.impl.xml.struct.LifeCycleCallback.PRE_PASSIVATE;
import org.ow2.util.ee.metadata.common.impl.xml.struct.MessageDestinationRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.PortComponentRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.ResourceEnvRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.ResourceRef;
import org.ow2.util.ee.metadata.common.impl.xml.struct.ServiceRef;
import org.ow2.util.xml.XMLSerializer;
import org.ow2.util.xml.XMLUtils;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import javax.xml.namespace.QName;
import java.io.IOException;
import java.io.StringWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

/**
 * Provide a method to analyze an xml environnement.
 * @author Florent Benoit
 */
public abstract class CommonDeploymentDescLoader extends AbsDeploymentDescLoader {

    /**
     * Logger.
     */
    // private static final Log LOGGER = LogFactory.getLog(CommonDeploymentDescLoader.class);

    /**
     * Utility class, no public constructor.
     * @param url the url that is pointing to the EJB Xml file.
     */
    protected CommonDeploymentDescLoader(final URL url) {
        super(url);
    }

    /**
     * Analyze the environment (could be a bean, servlet, etc).
     * @param envElement the dom element.
     * @param env the structure representing object.
     */
    protected void analyzeEnvironment(final Element envElement, final AbsEnvironment env) {
        analyzeLifeCycle(envElement, env, POST_CONSTRUCT);
        analyzeLifeCycle(envElement, env, PRE_DESTROY);
        analyzeEnvEnvtry(envElement, env);
        analyzeEJBRef(envElement, env);
        analyzeEJBLocalRef(envElement, env);
        analyzeResourceRef(envElement, env);
        analyzeResourceEnvRef(envElement, env);
        analyzeMessageDestinationRef(envElement, env);
        analyzeServiceRef(envElement, env);
    }

    /**
     * Analyze the ejb-ref/local-ref element and find the sub-elements.
     * @param refElement given XML element which contains ejb-ref-name,
     *        ejb-ref-type, etc.
     * @param ref the struct object representing the XML element.
     */
    private void analyzeCommonEJBRef(final Element refElement, final AbsEJBRef ref) {
        // ejb-ref-name
        String ejbRefName = XMLUtils.getStringValueElement(getXMLNS(), refElement, "ejb-ref-name");
        ref.setEjbRefName(ejbRefName);

        // ejb-ref-type
        String ejbRefType = XMLUtils.getStringValueElement(getXMLNS(), refElement, "ejb-ref-type");
        ref.setEjbRefType(ejbRefType);

        // mapped-name
        String mappedName = XMLUtils.getStringValueElement(getXMLNS(), refElement, "mapped-name");
        ref.setMappedName(mappedName);

        // ejb-link
        String ejbLink = XMLUtils.getStringValueElement(getXMLNS(), refElement, "ejb-link");
        ref.setEjbLink(ejbLink);

        // Analyze injection-target
        analyzeInjectionTarget(refElement, ref);
    }

    /**
     * Analyze the injection-target element and add them to the given resource
     * group object used by ejb-ref, ejb-local-ref, resource-ref, etc.
     * @param refElement given XML element which contains a resource element.
     * @param resourceGroup the struct object representing the XML element.
     */
    private void analyzeInjectionTarget(final Element refElement, final AbsResourceGroup resourceGroup) {
        // injection target
        NodeList injectionTargetList = refElement.getElementsByTagNameNS(getXMLNS(), "injection-target");
        for (int j = 0; j < injectionTargetList.getLength(); j++) {
            Element injectionTargetElement = (Element) injectionTargetList.item(j);

            // Build Object and add it
            InjectionTarget injectionTarget = new InjectionTarget();

            // Get class name
            String className = XMLUtils.getStringValueElement(getXMLNS(), injectionTargetElement, "injection-target-class");
            injectionTarget.setClassname(className);

            // Get target
            String targetName = XMLUtils.getStringValueElement(getXMLNS(), injectionTargetElement, "injection-target-name");
            injectionTarget.setTargetName(targetName);

            resourceGroup.addInjectionTarget(injectionTarget);
        }

    }

    /**
     * Analyze the environment element and find the ejb-ref element.
     * @param envElement given XML element which contains ejb-ref.
     * @param env the struct object representing the XML element.
     */
    private void analyzeEJBRef(final Element envElement, final AbsEnvironment env) {
        NodeList ejbRefList = envElement.getElementsByTagNameNS(getXMLNS(), EJBRef.NAME);

        // Loop on this list
        for (int i = 0; i < ejbRefList.getLength(); i++) {
            Element ejbRefElement = (Element) ejbRefList.item(i);
            EJBRef ejbRef = new EJBRef();

            analyzeCommonEJBRef(ejbRefElement, ejbRef);

            // remote
            String remote = XMLUtils.getStringValueElement(getXMLNS(), ejbRefElement, "remote");
            ejbRef.setRemote(remote);

            // home
            String home = XMLUtils.getStringValueElement(getXMLNS(), ejbRefElement, "home");
            ejbRef.setHome(home);

            // add ejb-ref in environment element
            env.addEJBRef(ejbRef);
        }
    }

    /**
     * Analyze the environment element and find the resource-ref element.
     * @param envElement given XML element which contains resource-ref.
     * @param env the struct object representing the XML element.
     */
    private void analyzeResourceRef(final Element envElement, final AbsEnvironment env) {
        NodeList resourceRefList = envElement.getElementsByTagNameNS(getXMLNS(), ResourceRef.NAME);

        // Loop on this list
        for (int i = 0; i < resourceRefList.getLength(); i++) {
            Element resourceRefElement = (Element) resourceRefList.item(i);
            ResourceRef resourceRef = new ResourceRef();

            // name
            String resRefName = XMLUtils.getStringValueElement(getXMLNS(), resourceRefElement, "res-ref-name");
            resourceRef.setResRefName(resRefName);

            // type
            String resRefType = XMLUtils.getStringValueElement(getXMLNS(), resourceRefElement, "res-type");
            resourceRef.setResRefType(resRefType);

            // mapped-name
            String mappedName = XMLUtils.getStringValueElement(getXMLNS(), resourceRefElement, "mapped-name");
            resourceRef.setMappedName(mappedName);

            // Analyze injection-target
            analyzeInjectionTarget(resourceRefElement, resourceRef);

            // add resource-ref in environment element
            env.addResourceRef(resourceRef);
        }
    }

    /**
     * Analyze the environment element and find the message-destination-ref
     * element.
     * @param envElement given XML element which contains resource-ref.
     * @param env the struct object representing the XML element.
     */
    private void analyzeMessageDestinationRef(final Element envElement, final AbsEnvironment env) {
        NodeList messageDestinationRefList = envElement.getElementsByTagNameNS(getXMLNS(), MessageDestinationRef.NAME);

        // Loop on this list
        for (int i = 0; i < messageDestinationRefList.getLength(); i++) {
            Element messageDestinationRefElement = (Element) messageDestinationRefList.item(i);
            MessageDestinationRef messageDestinationRef = new MessageDestinationRef();

            // name
            String name = XMLUtils
                    .getStringValueElement(getXMLNS(), messageDestinationRefElement, "message-destination-ref-name");
            messageDestinationRef.setName(name);

            // type
            String type = XMLUtils.getStringValueElement(getXMLNS(), messageDestinationRefElement, "message-destination-type");
            messageDestinationRef.setType(type);

            // usage
            String usage = XMLUtils.getStringValueElement(getXMLNS(), messageDestinationRefElement, "message-destination-usage");
            messageDestinationRef.setUsage(usage);

            // link
            String link = XMLUtils.getStringValueElement(getXMLNS(), messageDestinationRefElement, "message-destination-link");
            messageDestinationRef.setLink(link);

            // mapped-name
            String mappedName = XMLUtils.getStringValueElement(getXMLNS(), messageDestinationRefElement, "mapped-name");
            messageDestinationRef.setMappedName(mappedName);

            // Analyze injection-target
            analyzeInjectionTarget(messageDestinationRefElement, messageDestinationRef);

            // add message-destination-ref in environment element
            env.addMessageDestinationRef(messageDestinationRef);
        }
    }

    /**
     * Analyze the environment element and find the service-ref
     * element.
     * @param envElement given XML element which contains service-ref.
     * @param env the struct object representing the XML element.
     */
    private void analyzeServiceRef(final Element envElement, final AbsEnvironment env) {
        NodeList serviceRefList = envElement.getElementsByTagNameNS(getXMLNS(), ServiceRef.NAME);

        // Loop on this list
        for (int i = 0; i < serviceRefList.getLength(); i++) {
            Element serviceRefElement = (Element) serviceRefList.item(i);
            ServiceRef serviceRef = new ServiceRef();

            // name
            String name = XMLUtils
                    .getStringValueElement(getXMLNS(), serviceRefElement, "service-ref-name");
            serviceRef.setName(name);

            // service-interface
            String type = XMLUtils.getStringValueElement(getXMLNS(), serviceRefElement, "service-interface");
            serviceRef.setServiceInterface(type);

            // service-ref-type
            String usage = XMLUtils.getStringValueElement(getXMLNS(), serviceRefElement, "service-ref-type");
            serviceRef.setServiceRefType(usage);

            // wsdl-file
            String link = XMLUtils.getStringValueElement(getXMLNS(), serviceRefElement, "wsdl-file");
            URI location = null;
            try {
                location = new URI(link);
            } catch (URISyntaxException e) {
                throw new IllegalStateException("Element 'wsdl-file' is not a valid URI", e);
            }
            serviceRef.setWsdlFile(location);

            // service-qname
            NodeList serviceQNameList = envElement.getElementsByTagNameNS(getXMLNS(), "service-qname");
            if (serviceQNameList.getLength() == 1) {
                QName serviceQName = XMLUtils.getValueAsQName((Element) serviceQNameList.item(0));
                serviceRef.setServiceQName(serviceQName);
            }

            // port-component-ref
            analyzePortComponentRef(serviceRefElement, serviceRef);

            // handler-chains
            analyzeHandlerChains(serviceRefElement, serviceRef);

            // Analyze injection-target
            analyzeInjectionTarget(serviceRefElement, serviceRef);

            // add service-ref in environment element
            env.addServiceRef(serviceRef);
        }
    }

    /**
     * Analyze the service-ref element and find the handler(s) or handler-chains
     * element.
     * @param serviceRefElement given XML element which contains port-component-ref.
     * @param serviceRef the struct object representing the XML element.
     */
    private void analyzeHandlerChains(Element serviceRefElement, ServiceRef serviceRef) {
        NodeList handlerChainsList = serviceRefElement.getElementsByTagNameNS(getXMLNS(), "handler-chains");

        if (handlerChainsList.getLength() == 1) {
            // There is a handler-chains element
            Element handlerChainsElement = (Element) handlerChainsList.item(0);

            /*
             * Get all the required XML namespace used in the handler-chain:
             *  + soap-header: javaee:xsdQNameType
             *  + service-name-pattern, port-name-pattern: javaee:service-ref_qname-pattern
             */
            moveNamespacesDeclaration(handlerChainsElement, "soap-header");
            moveNamespacesDeclaration(handlerChainsElement, "service-name-pattern");
            moveNamespacesDeclaration(handlerChainsElement, "port-name-pattern");

            // Prepare serializer
            StringWriter writer = new StringWriter();
            XMLSerializer serializer = new XMLSerializer(handlerChainsElement);
            try {
                serializer.serialize(writer);
                StringBuffer sb = writer.getBuffer();

                // Store the serialized form in the ServiceRef XML structure
                serviceRef.setHandlerChains(sb.toString());
            } catch (IOException e) {
                throw new IllegalArgumentException("Unable to serialize handlers description for service-ref '"
                                                   + serviceRef.getName() + "'.", e);
            }

        }
    }

    private void moveNamespacesDeclaration(Element handlerChainsElement, String localName) {
        // Store service-ref element for later use
        Element serviceRefElement = (Element) handlerChainsElement.getParentNode();

        // Analyze all elements of the specified localName
        NodeList elementList = handlerChainsElement.getElementsByTagNameNS(getXMLNS(), localName);
        for (int i = 0; i < elementList.getLength(); i++) {
            Element element = (Element) elementList.item(i);

            // Extract QName
            QName name = XMLUtils.getValueAsQName(element);

            // Got the QName, now we find where it is declared
            // by traversing the ancestor hierarchy from child to parents
            Element declaringParent = null;

            // Start from service-ref
            Element traversed = serviceRefElement;
            Element root = serviceRefElement.getOwnerDocument().getDocumentElement();
            
            // Search until we find the declaring element or we traversed
            // all the hierarchy up to the document element.
            while ((declaringParent == null) && (!root.isEqualNode(traversed))) {
                Attr declaration = traversed.getAttributeNodeNS("http://www.w3.org/2000/xmlns/", name.getPrefix());
                if (declaration != null) {
                    declaringParent = traversed;
                } else {
                    traversed = (Element) element.getParentNode();
                }
            }

            if (declaringParent != null) {
                // OK, this namespace was declared in a non serialized element
                // We have to add it by hand in the handler-chains element
                // It's normal that theses new namespaces declaration overrides
                // the olds ones, but as they don't change the URI, that's fine

                handlerChainsElement.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:" + name.getPrefix(), name.getNamespaceURI());

            }
            // else {
            //   We did not find the namespace on nodes ancestors of handler-chains
            //   So it has been declared in an element that will be serialized
            // }
        }
    }

    /**
     * Analyze the service-ref element and find the port-component-ref
     * element.
     * @param serviceRefElement given XML element which contains port-component-ref.
     * @param serviceRef the struct object representing the XML element.
     */
    private void analyzePortComponentRef(final Element serviceRefElement, final ServiceRef serviceRef) {
        NodeList portComponentRefList = serviceRefElement.getElementsByTagNameNS(getXMLNS(), PortComponentRef.NAME);

        // Loop on this list
        for (int i = 0; i < portComponentRefList.getLength(); i++) {
            Element portComponentRefElement = (Element) portComponentRefList.item(i);
            PortComponentRef portComponentRef = new PortComponentRef();

            // service-endpoint-interface
            String sei = XMLUtils.getStringValueElement(getXMLNS(), portComponentRefElement, "service-endpoint-interface");
            portComponentRef.setServiceEndpointInterface(sei);

            // enable-mtom
            String enableMTOM = XMLUtils.getStringValueElement(getXMLNS(), portComponentRefElement, "enable-mtom");
            if (enableMTOM != null) {
                portComponentRef.setEnableMTOM(Boolean.parseBoolean(enableMTOM));
            }

            // port-component-link
            String link = XMLUtils.getStringValueElement(getXMLNS(), portComponentRefElement, "port-component-link");
            portComponentRef.setPortComponentLink(link);

            // add port-component-ref in service-ref element
            serviceRef.addPortComponentRef(portComponentRef);
        }
    }

    /**
     * Analyze the environment element and find the resource-env-ref element.
     * @param envElement given XML element which contains resource-ref.
     * @param env the struct object representing the XML element.
     */
    private void analyzeResourceEnvRef(final Element envElement, final AbsEnvironment env) {
        NodeList resourceEnvRefList = envElement.getElementsByTagNameNS(getXMLNS(), ResourceEnvRef.NAME);

        // Loop on this list
        for (int i = 0; i < resourceEnvRefList.getLength(); i++) {
            Element resourceEnvRefElement = (Element) resourceEnvRefList.item(i);
            ResourceEnvRef resourceEnvRef = new ResourceEnvRef();

            // name
            String resourceEnvRefName = XMLUtils
                    .getStringValueElement(getXMLNS(), resourceEnvRefElement, "resource-env-ref-name");
            resourceEnvRef.setResourceEnvRefName(resourceEnvRefName);

            // type
            String resourceEnvRefType = XMLUtils
                    .getStringValueElement(getXMLNS(), resourceEnvRefElement, "resource-env-ref-type");
            resourceEnvRef.setResourceEnvRefType(resourceEnvRefType);

            // mapped-name
            String mappedName = XMLUtils.getStringValueElement(getXMLNS(), resourceEnvRefElement, "mapped-name");
            resourceEnvRef.setMappedName(mappedName);

            // Analyze injection-target
            analyzeInjectionTarget(resourceEnvRefElement, resourceEnvRef);

            // add resource-ref in environment element
            env.addResourceEnvRef(resourceEnvRef);
        }
    }

    /**
     * Analyze the environment element and find the ejb-local-ref element.
     * @param envElement given XML element which contains ejb-local-ref.
     * @param env the struct object representing the XML element.
     */
    private void analyzeEJBLocalRef(final Element envElement, final AbsEnvironment env) {
        NodeList ejbLocalRefList = envElement.getElementsByTagNameNS(getXMLNS(), EJBLocalRef.NAME);

        // Loop on this list
        for (int i = 0; i < ejbLocalRefList.getLength(); i++) {
            Element ejbRefElement = (Element) ejbLocalRefList.item(i);
            EJBLocalRef ejbLocalRef = new EJBLocalRef();

            analyzeCommonEJBRef(ejbRefElement, ejbLocalRef);

            // local
            String local = XMLUtils.getStringValueElement(getXMLNS(), ejbRefElement, "local");
            ejbLocalRef.setLocal(local);

            // local-home
            String localHome = XMLUtils.getStringValueElement(getXMLNS(), ejbRefElement, "local-home");
            ejbLocalRef.setLocalHome(localHome);

            // add ejb-local-ref in environment element
            env.addEJBLocalRef(ejbLocalRef);
        }
    }

    /**
     * Analyze the env-entry element.
     * @param envElement the dom element.
     * @param bean the structure representing object.
     */
    private void analyzeEnvEnvtry(final Element envElement, final AbsEnvironment bean) {
        // Get env-entry
        NodeList envEntryList = envElement.getElementsByTagNameNS(getXMLNS(), EnvEntry.NAME);

        // Loop on env entry
        for (int i = 0; i < envEntryList.getLength(); i++) {
            Element envEntryElement = (Element) envEntryList.item(i);

            // Build instance that is created.
            EnvEntry envEntry = new EnvEntry();
            bean.addEnvEntry(envEntry);

            // Env-entry name
            String envEntryName = XMLUtils.getStringValueElement(getXMLNS(), envEntryElement, "env-entry-name");
            envEntry.setEnvEntryName(envEntryName);

            // Env-entry type
            String envEntryType = XMLUtils.getStringValueElement(getXMLNS(), envEntryElement, "env-entry-type");
            envEntry.setEnvEntryType(envEntryType);

            // Env-entry value
            String envEntryValue = XMLUtils.getStringValueElement(getXMLNS(), envEntryElement, "env-entry-value");
            envEntry.setEnvEntryValue(envEntryValue);

            // Analyze injection-target
            analyzeInjectionTarget(envEntryElement, envEntry);

        }
    }

    /**
     * Analyze the lifecycle elements.
     * @param envElement the dom element.
     * @param env the structure representing object.
     * @param lifecycleName the name of lifecycle (post-construct, etc)
     */
    protected void analyzeLifeCycle(final Element envElement, final AbsEnvironment env, final String lifecycleName) {
        // Get the lifecycle elements
        NodeList lifecycleList = envElement.getElementsByTagNameNS(getXMLNS(), lifecycleName);

        // Loop on the lifecycle
        for (int i = 0; i < lifecycleList.getLength(); i++) {
            Element lifeCycleElement = (Element) lifecycleList.item(i);

            // Build instance that is created.
            LifeCycleCallback lifecycleCallback = new LifeCycleCallback();
            if (POST_CONSTRUCT.equals(lifecycleName)) {
                env.addPostConstructCallback(lifecycleCallback);
            } else if (PRE_DESTROY.equals(lifecycleName)) {
                env.addPreDestroyCallback(lifecycleCallback);
            } else if (PRE_PASSIVATE.equals(lifecycleName)) {
                env.addPrePassivateCallback(lifecycleCallback);
            } else if (POST_ACTIVATE.equals(lifecycleName)) {
                env.addPostActivateCallback(lifecycleCallback);
            } else {
                throw new IllegalArgumentException("Unknown lifecycle named '" + lifecycleName + "'.");
            }

            // lifecycle-callback-class
            String className = XMLUtils.getStringValueElement(getXMLNS(), lifeCycleElement, "lifecycle-callback-class");
            lifecycleCallback.setLifecycleCallbackClass(className);

            // lifecycle-callback-method
            String method = XMLUtils.getStringValueElement(getXMLNS(), lifeCycleElement, "lifecycle-callback-method");
            lifecycleCallback.setMethodName(method);

        }

    }


}
